﻿class RecordSet extends RsDataProviderClass {
	private var mTitles = null;
	private var serverinfo = null;
	private var mRecordsAvailable = null;
	private var mRecordSetId = null;
	private var serviceName = null;
	private var mTotalCount = null;
	private var mDeliveryMode = null;
	private var mAllNotified = null;
	private var mOutstandingRecordCount = null;
	private var gateway_conn = null;
	private var mRecordSetService = null;
	private var mRecordSetID = null;
	private var mPageSize = null;
	private var mNumPrefetchPages = null;
	private var recordIndex = null;
	private var mDataFetcher = null;
	function RecordSet(columnNames) {
		if (this.mTitles != null) {
			// we've already been constructed.
			// this should happen only when the object is a
			// shared object that's being reloaded into memory.
			this.views = new Array();
			// these are probably out of date
			return;
		}
		// initialize our superclass
		this.init();
		if (this.serverinfo == null) {
			// there's no server information - we're done
			this.mTitles = columnNames;
			return;
		}
		// this recordset came from a server.
		if (this.serverinfo.version != 1) {
			NetServices.trace("RecordSet", "warning", 100, "Received incompatible recordset version from server");
			return;
		}
		// set up the initial contents of the recordset
		this.mTitles = this.serverinfo.columnNames;
		this.mRecordsAvailable = 0;
		//this.mHighWater = 0;
		this.setData((this.serverinfo.cursor == null) ? 0 : (this.serverinfo.cursor-1), this.serverinfo.initialData);
		if (this.serverinfo.initialData.length != this.serverinfo.totalCount) {
			this.mRecordSetId = this.serverinfo.id;
			// we haven't yet got all the data
			if (this.mRecordSetId != null) {
				// if id is non-null, there are more records still on the server.
				// This therefore is a server-associated recordset
				this.serviceName = (this.serverinfo.serviceName == null) ? "RecordSet" : this.serverinfo.serviceName;
				// initialize other fields
				this.mTotalCount = this.serverinfo.totalCount;
				this.mDeliveryMode = "ondemand";
				this.mAllNotified = false;
				this.mOutstandingRecordCount = 0;
			} else {
				NetServices.trace("RecordSet", "warning", 102, "Missing some records, but there's no recordset id");
			}
		}
		this.serverinfo = null;
	}
	public function _setParentService(service) {
		this.gateway_conn = service.nc;
	}
	public function getRecordSetService() {
		if (this.mRecordSetService == null) {
			if (this.gateway_conn == null) {
				// _setParentService never got called. this will only happen if:
				//  - recordset is not the top-level return value, or
				//  - NetServiceProxyResponder is not in use (i.e. there is a developer-supplied response object 
				//    for the service call that returned this RecordSet
				this.gateway_conn = NetServices.createGatewayConnection();
			} else {
				// A parent netconnect has been supplied. If necessary,
				// make a new netconnection, so we can have separate debug flags
				if (_global.netDebugInstance != undefined) {
					this.gateway_conn = this.gateway_conn.clone();
				}
			}
			if (_global.netDebugInstance != undefined) {
				this.gateway_conn.setupRecordset();
				this.gateway_conn.setDebugId("RecordSet "+this.mRecordSetId);
			}
			this.mRecordSetService = this.gateway_conn.getService(this.serviceName, this);
			if (this.mRecordSetService == null) {
				NetServices.trace("RecordSet", "warning", 101, "Failed to create recordset service");
				this.mRecordSetService = 0;
				// so we don't get a flood of error messages
			}
		}
		return this.mRecordSetService;
	}
	//-------------------------------------------------------------
	// public property access functions
	//-------------------------------------------------------------
	public function getColumnNames() {
		return this.mTitles;
	}
	public function getLength() {
		if (this.mRecordSetID != null) {
			// a server-associated recordset
			//trace("RecordSet.getLength: " + this.mTotalCount);
			return this.mTotalCount;
		} else {
			// a local recordset
			//trace("RecordSet.getLength: " + this.items.length);
			return this.items.length;
		}
	}
	// returns null if the index is out of range
	// returns 1 if the data is unavailable
	// returns the data if it is available
	public function getItemAt(index) {
		//trace("RecordSet.getItemAt(" + index + ")");
		// It is server associated. See if the index is valid. 
		if ((index<0) || (index>=this.getLength())) {
			NetServices.trace("RecordSet", "warning", 104, "getItemAt("+index+") index out of range");
			return null;
		}
		if (this.mRecordSetId == null) {
			// this is not a server-associated recordset
			//trace("RecordSet.getItemAt not server-associated");
			return this.items[index];
		}
		// let the paging lookahead code have a look at this request.
		// (even if the record is in memory we might still want to request a download of the page)
		this.requestRecord(index);
		var result = this.items[index];
		if (result == 1) {
			return "in progress";
		}
		//trace("RecordSet.getItemAt(" + index + ") returning " + result);
		return result;
	}
	// the spec includes this as a dataprovider function, even though it really isn't
	public function setField(index, fieldName, value) {
		if (!this.checkLocal()) {
			return;
		}
		if (index<0 || index>=this.getLength()) {
			return;
		}
		this.items[index][fieldName] = value;
		this.updateViews({event:"updateRows", firstRow:index, lastRow:index});
	}
	//-------------------------------------------------------------
	// public Data Manipulation functions
	//-------------------------------------------------------------
	public function filter(filterFunction, context) {
		if (!this.checkLocal()) {
			return;
		}
		// create a new, empty recordset
		var result = new RecordSet(this.mTitles);
		// loop over all the records in the current recordset
		// find the ones that the the filter function approves of,
		// and add it to the result
		var rcount = this.getLength();
		for (var i = 0; i<rcount; i++) {
			var item = this.getItemAt(i);
			if ((item != null) && (item != 1) && filterFunction(item, context)) {
				result.addItem(item);
			}
		}
		//trace("RecordSet.filter: " + NetServices.objectToString(result));
		return result;
	}
	public function sort(compareFunc) {
		if (!this.checkLocal()) {
			return;
		}
		this.items.sort(compareFunc);
		this.updateViews({event:"sort"});
	}
	//-------------------------------------------------------------
	// public Data Delivery functions
	//-------------------------------------------------------------
	public function isLocal() {
		return (this.mRecordSetID == null);
	}
	public function isFullyPopulated() {
		return this.isLocal();
	}
	public function getNumberAvailable() {
		if (this.isLocal()) {
			return this.getLength();
		} else {
			return this.mRecordsAvailable;
		}
	}
	public function setDeliveryMode(mode, pagesize, numPrefetchPages) {
		//trace("RecordSet.setDeliveryMode(" + mode + "," + pagesize + "," + numPrefetchPages + ")");
		this.mDeliveryMode = mode;
		this.stopFetchAll();
		if (mode == "ondemand") {
			// nothing else to do
			return;
		}
		if (pagesize == null) {
			// deduce the page size from our views...
			// !!@ doc note: you must up at least one view (via setDataProvider) before doing setDeliveryMode
			// !!@ doc note: if there are multiple views, we use the pagesize of the first view 
			// !!@ doc note: if the listbox gets resized, you must call setDeliveryMode again
			pagesize = this.views[0].getRowCount();
			if (pagesize == null) {
				pagesize = 25;
			}
		}
		if (mode == "page") {
			if (numPrefetchPages == null) {
				numPrefetchPages = 0;
			}
			this.mPageSize = pagesize;
			this.mNumPrefetchPages = numPrefetchPages;
			// !!@ doc note: 0 is ok for numPrefetchPages, it means fetch only the current page.
		} else if (mode == "fetchall") {
			this.stopFetchAll();
			this.startFetchAll(pagesize);
		} else {
			NetServices.trace("RecordSet", "warning", 107, "SetDeliveryMode: unknown mode string");
		}
	}
	//-------------------------------------------------------------
	// Responses from the RecordSet service
	//-------------------------------------------------------------
	// This function is where all our data comes in from the recordset service
	public function getRecords_Result(info) {
		//trace("RecordSet.getRecords_Result(), start=" + info.Cursor +
		//	", id=" + info.id + ", data=" + info.Page);
		this.setData(info.Cursor-1, info.Page);
		this.mOutstandingRecordCount -= info.Page.length;
		// !!@ assert (this.mOutstandingRecordCount) >= 0
		this.updateViews({event:"updateRows", firstRow:info.Cursor-1, lastRow:info.Cursor-1+info.Page.length-1});
		if ((this.mRecordsAvailable == this.mTotalCount) && !this.mAllNotified) {
			this.updateViews({event:"allRows"});
			this.mRecordSetService.release();
			this.mAllNotified = true;
			this.mRecordSetID = null;
			this.mRecordSetService = null;
		}
	}
	public function release_Result() {
		// ignore this
	}
	//-------------------------------------------------------------
	// private data delivery functions
	//-------------------------------------------------------------
	public function arrayToObject(anArray) {
		if (this.mTitles == null) {
			NetServices.trace("RecordSet", "warning", 105, "getItem: titles are not available");
			return null;
		}
		var result = new Object();
		var alen = anArray.length;
		for (var i = 0; i<alen; i++) {
			var title = this.mTitles[i];
			if (title == null) {
				title = "column"+i+1;
			}
			result[title] = anArray[i];
		}
		return result;
	}
	public function setData(start, dataArray) {
		//trace("setData " + start + "," +dataArray.length);
		var datalen = dataArray.length;
		for (var i = 0; i<datalen; i++) {
			var index = i+start;
			var rec = this.items[index];
			if ((rec != null) && (rec != 1)) {
				// why are we getting this data! we already have it
				NetServices.trace("RecordSet", "warning", 106, "Already got record # "+recordIndex);
			} else {
				this.mRecordsAvailable += 1;
			}
			this.items[index] = this.arrayToObject(dataArray[i]);
			this.items[index].__ID__ = this.uniqueID++;
		}
	}
	public function requestOneRecord(index) {
		if (this.items[index] == null) {
			this.getRecordSetService().getRecords(this.mRecordSetId, index+1, 1);
			this.mOutstandingRecordCount++;
			this.items[index] = 1;
			this.updateViews({event:"fetchRows", firstRow:index, lastRow:index});
		}
	}
	public function requestRecord(index) {
		//trace("RecordSet.requestRecord(" + index + ")");
		if (this.mDeliveryMode != "page") {
			// fetchall, ondemand or unknown
			this.requestOneRecord(index);
			return;
			// !!@ if in fetchall mode, should start fetching from here, and wrap around to 1 later on
		}
		// we're in page mode
		// !!@ there is probably a better algorithm, but messier
		// the current algorithm works fine if records are always fetched in increasing index order
		// but goes wierd if not.
		// See if we need to prefetch the next N+1 pages, starting at index
		// (its N+1 because we always prefetch *the current page* as well
		// as any requested other pages)
		var firstIndex = int(index/this.mPageSize)*this.mPageSize;
		var lastIndex = firstIndex+(this.mPageSize*(this.mNumPrefetchPages+1))-1;
		this.requestRecordRange(firstIndex, lastIndex);
	}
	public function requestRecordRange(index, lastIndex) {
		//trace("RecordSet.requestRecordRange(" + index + "," + lastIndex + ")");
		var highestRequested = -1;
		// make sure indices are valid
		if (index<0) {
			index = 0;
		}
		if (lastIndex>=this.getLength()) {
			lastIndex = this.getLength()-1;
		}
		// find sequences of null entries to request
		// we could also just make a bunch of individual requests, but this seems 
		// cleaner. The server should be able to handle either case efficiently.
		while (index<=lastIndex) {
			while ((index<=lastIndex) && (this.items[index] != null)) {
				index++;
			}
			var first = index;
			// this is the index of the first null entry in a group
			while ((index<=lastIndex) && (this.items[index] == null)) {
				this.mOutstandingRecordCount++;
				this.items[index] = 1;
				index++;
			}
			var last = index-1;
			if (first<=last) {
				this.getRecordSetService().getRecords(this.mRecordSetId, first+1, last-first+1);
				highestRequested = last;
				this.updateViews({event:"fetchRows", firstRow:first, lastRow:last});
			}
		}
		return highestRequested;
	}
	public function startFetchAll(pagesize) {
		//trace("RecordSet.startFetchAll()");
		this.mDataFetcher.disable();
		this.mDataFetcher = new RsDataFetcher(this, pagesize);
	}
	public function stopFetchAll() {
		//trace("RecordSet.stopFetchAll()");
		this.mDataFetcher.disable();
		this.mDataFetcher = null;
	}
	public function checkLocal() {
		if (this.isLocal()) {
			return true;
		} else {
			NetServices.trace("RecordSet", "warning", 108, "Operation not allowed on partial recordset");
			return false;
		}
	}
}
